/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package explorviz.hpc_monitoring.reader;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.zip.Inflater;
import kieker.analysis.IProjectContext;
import kieker.analysis.plugin.annotation.*;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.common.configuration.Configuration;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.IMonitoringRecord;
import com.rabbitmq.client.*;
import explorviz.hpc_monitoring.Bits;
import explorviz.hpc_monitoring.record.Trace;
import explorviz.hpc_monitoring.record.events.*;

/**
 * Reads monitoring records from the queue of an established RabbitMQ
 * connection.
 * 
 * @author Santje Finke
 * 
 * @since 1.8
 */
@Plugin(description = "A reader which reads records from a RabbitMQ queue", dependencies = "This plugin needs the file 'rabbitmq.client-*.jar'.", outputPorts = { @OutputPort(name = RabbitMQReader.OUTPUT_PORT_NAME_RECORDS, eventTypes = { IMonitoringRecord.class }, description = "Output Port of the JMSReader") }, configuration = {
        @Property(name = RabbitMQReader.CONFIG_PROPERTY_NAME_PROVIDERURL, defaultValue = "localhost"),
        @Property(name = RabbitMQReader.CONFIG_PROPERTY_NAME_QUEUE, defaultValue = "kieker"),
        @Property(name = RabbitMQReader.CONFIG_PROPERTY_PASSWORD, defaultValue = "guest"),
        @Property(name = RabbitMQReader.CONFIG_PROPERTY_USER, defaultValue = "guest"),
        @Property(name = RabbitMQReader.CONFIG_PROPERTY_PORT, defaultValue = "5672")

})
public final class RabbitMQReader extends AbstractReaderPlugin {

    /** The name of the output port delivering the received records. */
    public static final String             OUTPUT_PORT_NAME_RECORDS         = "monitoringRecords";
    /** The name of the configuration determining the RabbitMQ provider URL. */
    public static final String             CONFIG_PROPERTY_NAME_PROVIDERURL = "mqProviderUrl";
    /**
     * The name of the configuration determining the RabbitMQ Queue (e.g.
     * queue1).
     */
    public static final String             CONFIG_PROPERTY_NAME_QUEUE       = "mqDestination";
    /** The username that is used to connect to a queue. */
    public static final String             CONFIG_PROPERTY_USER             = "guest";
    /** The password that is used to connect to a queue. */
    public static final String             CONFIG_PROPERTY_PASSWORD         = "guest";
    /** The port that is used to connect to a queue. */
    public static final String             CONFIG_PROPERTY_PORT             = "5672";

    static final Log                       LOG                              = LogFactory
                                                                                    .getLog(RabbitMQReader.class); // NOPMD
                                                                                                                   // package
                                                                                                                   // for
                                                                                                                   // inner
                                                                                                                   // class

    private final String                   providerUrl;
    private final String                   queueName;
    private final String                   password;
    private final String                   username;
    private final int                      port;
    private Connection                     connection;
    private Channel                        channel;
    private ConnectionFactory              factory;
    private QueueingConsumer               normalConsumer;

    private final CountDownLatch           cdLatch                          = new CountDownLatch(
                                                                                    1);

    private final Map<Integer, String>     stringRegistry                   = new ConcurrentHashMap<Integer, String>(
                                                                                    16,
                                                                                    0.75f,
                                                                                    2);

    private final RabbitMQRegistryConsumer registryConsumer;

    /**
     * Creates a new instance of this class using the given parameters.
     * 
     * @param configuration
     *            The configuration used to initialize the whole reader. Keep in
     *            mind that the configuration should contain the following
     *            properties:
     *            <ul>
     *            <li>The property {@link #CONFIG_PROPERTY_NAME_PROVIDERURL},
     *            e.g. {@code localhost}
     *            <li>The property {@link #CONFIG_PROPERTY_NAME_QUEUE}, e.g.
     *            {@code queue1}
     *            <li>The property {@link #CONFIG_PROPERTY_PASSWORD}, e.g.
     *            {@code password}
     *            <li>The property {@link #CONFIG_PROPERTY_USER}, e.g.
     *            {@code username}
     *            <li>The property {@link #CONFIG_PROPERTY_PORT}, e.g.
     *            {@code port}
     *            </ul>
     * @param projectContext
     *            The project context for this component.
     * 
     * @throws IllegalArgumentException
     *             If one of the properties is empty.
     */
    public RabbitMQReader(final Configuration configuration,
            final IProjectContext projectContext)
            throws IllegalArgumentException {
        super(configuration, projectContext);

        // Initialize the reader bases on the given configuration.
        providerUrl = configuration
                .getStringProperty(CONFIG_PROPERTY_NAME_PROVIDERURL);
        queueName = configuration.getStringProperty(CONFIG_PROPERTY_NAME_QUEUE);
        username = configuration.getStringProperty(CONFIG_PROPERTY_USER);
        password = configuration.getStringProperty(CONFIG_PROPERTY_PASSWORD);
        port = configuration.getIntProperty(CONFIG_PROPERTY_PORT);
        // simple sanity check
        if ((providerUrl.length() == 0) || (queueName.length() == 0)) {
            throw new IllegalArgumentException(
                    "RabbitMQReader has not sufficient parameters. providerUrl ('"
                            + providerUrl + "') or destination ('" + queueName
                            + "'), is null");
        }

        registryConsumer = new RabbitMQRegistryConsumer(this, providerUrl,
                "registryRecords", username, password, port);
        registryConsumer.start();
    }

    /**
     * Creates a new instance of this class using the given parameters.
     * 
     * @param configuration
     *            The configuration used to initialize the whole reader. Keep in
     *            mind that the configuration should contain the following
     *            properties:
     *            <ul>
     *            <li>The property {@link #CONFIG_PROPERTY_NAME_PROVIDERURL},
     *            e.g. {@code localhost}
     *            <li>The property {@link #CONFIG_PROPERTY_NAME_QUEUE}, e.g.
     *            {@code queue1}
     *            <li>The property {@link #CONFIG_PROPERTY_PASSWORD}, e.g.
     *            {@code password}
     *            <li>The property {@link #CONFIG_PROPERTY_USER}, e.g.
     *            {@code username}
     *            </ul>
     * 
     * @throws IllegalArgumentException
     *             If one of the properties is empty.
     * 
     * @deprecated To be removed in Kieker 1.8.
     */
    @Deprecated
    public RabbitMQReader(final Configuration configuration)
            throws IllegalArgumentException {
        this(configuration, null);
    }

    /**
     * A call to this method is a blocking call.
     * 
     * @return true if the method succeeds, false otherwise.
     */
    public boolean read() {
        boolean retVal = true;
        try {
            createConnectionFactory();
            connect();
            while (!Thread.interrupted()) {

                if (!connection.isOpen() || !channel.isOpen()) {
                    reconnect();
                }

                final QueueingConsumer.Delivery delivery = normalConsumer
                        .nextDelivery();

                byte[] batchedMessages = delivery.getBody();
                messagesfromByteArray(batchedMessages);

                normalConsumer.getChannel().basicAck(
                        delivery.getEnvelope().getDeliveryTag(), false);
            }
        }
        catch (final IOException ex) { // NOPMD NOCS (IllegalCatchCheck)
            LOG.error("Error in read()", ex);
            retVal = false;
        }
        catch (final ShutdownSignalException e) {
            LOG.error("Error in read(): ShutdownSignal Exception", e);
            retVal = false;
        }
        catch (final ConsumerCancelledException e) {
            LOG.error("Error in read(): ConsumerCancelled Exception", e);
            retVal = false;
        }
        catch (final InterruptedException e) {
            LOG.error("Error in read(): Interrupted Exception", e);
            retVal = false;
        }
        finally {
            try {
                if (connection != null) {
                    connection.close();
                }
            }
            catch (final IOException e) {
                LOG.error("Error in read()", e);
            }
        }
        return retVal;
    }

    public byte[] decompressByteArray(byte[] bytes) {

        ByteArrayOutputStream baos = null;
        Inflater iflr = new Inflater();
        iflr.setInput(bytes);
        baos = new ByteArrayOutputStream();
        byte[] tmp = new byte[4096];
        try {
            while (!iflr.finished()) {
                int size = iflr.inflate(tmp);
                baos.write(tmp, 0, size);
            }
        }
        catch (Exception ex) {

        }
        finally {
            try {
                if (baos != null) {
                    baos.close();
                }
            }
            catch (Exception ex) {}
        }

        return baos.toByteArray();
    }

    private void createConnectionFactory() {
        factory = new ConnectionFactory();
        factory.setHost(providerUrl);
        factory.setPort(port);
        factory.setConnectionTimeout(0);
        factory.setUsername(username);
        factory.setPassword(password);
    }

    private void connect() throws IOException {
        connection = factory.newConnection();
        channel = connection.createChannel();

        channel.queueDeclare(queueName, false, false, false, null);

        normalConsumer = new QueueingConsumer(channel);
        channel.basicConsume(queueName, false, normalConsumer);
        channel.basicQos(50);
    }

    private void reconnect() {
        try {
            connect();
        }
        catch (final IOException e) {
            RabbitMQReader.LOG.error("Error reestablishing connection", e);
        }
    }

    private void messagesfromByteArray(final byte[] b) {
        int offset = 0;

        int firstValue;
        while ((firstValue = Bits.getInt(b, offset)) != -1) {
            offset += 4;

            IMonitoringRecord record = null;

            switch (firstValue) {
                case 0: {
                    final long traceId = Bits.getLong(b, offset);
                    offset += 8;
                    final Integer hostnameId = Bits.getInt(b, offset);
                    offset += 4;
                    final long parentTraceId = Bits.getLong(b, offset);
                    offset += 8;
                    final int parentOrderId = Bits.getInt(b, offset);
                    offset += 4;
                    offset += 4; // dummy for nextOrderIndex

                    record = new Trace(traceId,
                            getStringFromRegistry(hostnameId), parentTraceId,
                            parentOrderId);
                    break;
                }
                case 1: {
                    final long timestamp = Bits.getLong(b, offset);
                    offset += 8;
                    final long traceId = Bits.getLong(b, offset);
                    offset += 8;
                    final int orderIndex = Bits.getInt(b, offset);
                    offset += 4;
                    final Integer operationId = Bits.getInt(b, offset);
                    offset += 4;

                    record = new BeforeOperationEvent(timestamp, traceId,
                            orderIndex, getStringFromRegistry(operationId));
                    break;
                }
                case 2: {
                    final long timestamp = Bits.getLong(b, offset);
                    offset += 8;
                    final long traceId = Bits.getLong(b, offset);
                    offset += 8;
                    final int orderIndex = Bits.getInt(b, offset);
                    offset += 4;
                    final Integer operationId = Bits.getInt(b, offset);
                    offset += 4;
                    final Integer causeId = Bits.getInt(b, offset);
                    offset += 4;

                    record = new AfterFailedOperationEvent(timestamp, traceId,
                            orderIndex, getStringFromRegistry(operationId),
                            getStringFromRegistry(causeId));
                    break;
                }
                case 3: {
                    final long timestamp = Bits.getLong(b, offset);
                    offset += 8;
                    final long traceId = Bits.getLong(b, offset);
                    offset += 8;
                    final int orderIndex = Bits.getInt(b, offset);
                    offset += 4;
                    final Integer operationId = Bits.getInt(b, offset);
                    offset += 4;

                    record = new AfterOperationEvent(timestamp, traceId,
                            orderIndex, getStringFromRegistry(operationId));
                    break;
                }
                default: {
                    LOG.error("unknown class id " + firstValue);
                }
            }

            if (!super.deliver(OUTPUT_PORT_NAME_RECORDS, record)) {
                LOG.error("deliverRecord returned false");
            }
        }
    }

    final void unblock() { // NOPMD (package visible for inner class)
        cdLatch.countDown();
    }

    /**
     * {@inheritDoc}
     */
    public void terminate(final boolean error) {
        LOG.info("Shutdown of RabbitMQReader requested.");
        registryConsumer.interrupt();
        unblock();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Configuration getCurrentConfiguration() {
        final Configuration configuration = new Configuration();

        configuration
                .setProperty(CONFIG_PROPERTY_NAME_PROVIDERURL, providerUrl);
        configuration.setProperty(CONFIG_PROPERTY_NAME_QUEUE, queueName);
        configuration.setProperty(CONFIG_PROPERTY_USER, username);
        configuration.setProperty(CONFIG_PROPERTY_PASSWORD, password);

        return configuration;
    }

    public void addToRegistry(Integer key, String value) {
        stringRegistry.put(key, value);
        System.out.println(key + " " + value);

        synchronized (this) {
            notifyAll();
        }
    }

    private String getStringFromRegistry(Integer id) {
        String result = stringRegistry.get(id);
        while (result == null) {
            try {
                synchronized (this) {
                    System.out.println("waiting for " + id);
                    this.wait();
                }
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }
            result = stringRegistry.get(id);
        }

        return result;
    }
}
