package explorviz.hpc_monitoring.reader;

import java.io.*;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import com.rabbitmq.client.*;
import explorviz.hpc_monitoring.StringRegistryRecord;

public class RabbitMQRegistryConsumer extends Thread {

    static final Log             LOG = LogFactory.getLog(RabbitMQReader.class); // NOPMD
                                                                                // package
                                                                                // for
                                                                                // inner
                                                                                // class

    private final RabbitMQReader parent;

    private Connection           connection;
    private Channel              channel;
    private ConnectionFactory    factory;
    private QueueingConsumer     registryConsumer;

    private final String         providerUrl;
    private final String         queueName;
    private final String         password;
    private final String         username;
    private final int            port;

    public RabbitMQRegistryConsumer(RabbitMQReader parent, String providerUrl,
            String queueName, String password, String username, int port) {
        this.parent = parent;
        this.providerUrl = providerUrl;
        this.queueName = queueName;
        this.password = password;
        this.username = username;
        this.port = port;

    }

    @Override
    public void run() {
        try {
            createConnectionFactory();

            connect();

            while (!Thread.interrupted()) {

                if (!connection.isOpen() || !channel.isOpen()) {
                    reconnect();
                }

                final QueueingConsumer.Delivery delivery = registryConsumer
                        .nextDelivery();

                final Object message = readObjectFromBytes(delivery.getBody());

                if (message instanceof StringRegistryRecord) {
                    final StringRegistryRecord record = (StringRegistryRecord) message;
                    parent.addToRegistry(record.getId(), record.getObject());
                }
            }
        }
        catch (final IOException ex) { // NOPMD NOCS (IllegalCatchCheck)
            LOG.error("Error in read()", ex);
        }
        catch (final ClassNotFoundException e) {
            LOG.error("Error in read(): ClassNotFound Exception", e);
        }
        catch (final ShutdownSignalException e) {
            LOG.error("Error in read(): ShutdownSignal Exception", e);
        }
        catch (final ConsumerCancelledException e) {
            LOG.error("Error in read(): ConsumerCancelled Exception", e);
        }
        catch (final InterruptedException e) {
            LOG.error("Error in read(): Interrupted Exception", e);
        }
        finally {
            try {
                if (connection != null) {
                    connection.close();
                }
            }
            catch (final IOException e) {
                LOG.error("Error in read()", e);
            }
        }
    }

    private void createConnectionFactory() {
        factory = new ConnectionFactory();
        factory.setHost(providerUrl);
        factory.setPort(port);
        factory.setConnectionTimeout(0);
        factory.setUsername(username);
        factory.setPassword(password);
    }

    private void connect() throws IOException {
        connection = factory.newConnection();
        channel = connection.createChannel();

        channel.queueDeclare(queueName, false, false, false, null);

        registryConsumer = new QueueingConsumer(channel);
        channel.basicConsume(queueName, true, registryConsumer);
    }

    private void reconnect() {
        try {
            connect();
        }
        catch (final IOException e) {
            RabbitMQReader.LOG.error("Error reestablishing connection", e);
        }
    }

    private Object readObjectFromBytes(final byte[] bs) throws IOException,
            ClassNotFoundException {
        final ByteArrayInputStream bain = new ByteArrayInputStream(bs);
        final ObjectInputStream in = new ObjectInputStream(bain);
        final Object message = in.readObject();
        return message;
    }
}
