package kieker.analysis.plugin.reader.mq;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.common.record.misc.RegistryRecord;

import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.rabbitmq.client.ConsumerCancelledException;
import com.rabbitmq.client.QueueingConsumer;
import com.rabbitmq.client.ShutdownSignalException;

public class RabbitMQRegistryConsumer extends Thread {
	
	static final Log LOG = LogFactory.getLog(RabbitMQReader.class); // NOPMD package for inner class
	
	private final RabbitMQReader parent;
	
	private Connection connection;
	private Channel channel;
	private ConnectionFactory factory;
	private QueueingConsumer registryConsumer;

	private String providerUrl;
	private String queueName;
	private String password;
	private String username;
	private int port;

	public RabbitMQRegistryConsumer(RabbitMQReader parent, String providerUrl, String queueName, String password, String username, int port) {
		this.parent = parent;
		this.providerUrl = providerUrl;
		this.queueName = queueName;
		this.password = password;
		this.username = username;
		this.port = port;
		
	}
	
	@Override
	public void run()  {
		try {
			createConnectionFactory();
			
			connect();
			
			while (!Thread.interrupted()) {

				if (!this.connection.isOpen() || !this.channel.isOpen()) {
					this.reconnect();
				}

				final QueueingConsumer.Delivery delivery = this.registryConsumer.nextDelivery();

				final Object message = readObjectFromBytes(delivery.getBody());
				
				
				if (message instanceof RegistryRecord) {
					@SuppressWarnings("unchecked")
					final  RegistryRecord<String> record = (RegistryRecord<String>) message;
					parent.addToRegistry(record.getId(), record.getObject());
				}
			}
		} catch (final IOException ex) { // NOPMD NOCS (IllegalCatchCheck)
			LOG.error("Error in read()", ex);
		} catch (final ClassNotFoundException e) {
			LOG.error("Error in read(): ClassNotFound Exception", e);
		} catch (final ShutdownSignalException e) {
			LOG.error("Error in read(): ShutdownSignal Exception", e);
		} catch (final ConsumerCancelledException e) {
			LOG.error("Error in read(): ConsumerCancelled Exception", e);
		} catch (final InterruptedException e) {
			LOG.error("Error in read(): Interrupted Exception", e);
		} finally {
			try {
				if (this.connection != null) {
					this.connection.close();
				}
			} catch (final IOException e) {
				LOG.error("Error in read()", e);
			}
		}
	}

	private void createConnectionFactory() {
		this.factory = new ConnectionFactory();
		this.factory.setHost(this.providerUrl);
		this.factory.setPort(this.port);
		this.factory.setConnectionTimeout(0);
		this.factory.setUsername(this.username);
		this.factory.setPassword(this.password);
	}
	
	private void connect() throws IOException {
		this.connection = this.factory.newConnection();
		this.channel = this.connection.createChannel();

		this.channel.queueDeclare(this.queueName, false, false, false, null);

		this.registryConsumer = new QueueingConsumer(this.channel);
		this.channel.basicConsume(this.queueName, true, this.registryConsumer);
	}
	
	private void reconnect() {
		try {
			connect();
		} catch (final IOException e) {
			RabbitMQReader.LOG.error("Error reestablishing connection", e);
		}
	}
	
	private Object readObjectFromBytes(final byte[] bs)
			throws IOException, ClassNotFoundException {
		final ByteArrayInputStream bain = new ByteArrayInputStream(bs);
		final ObjectInputStream in = new ObjectInputStream(bain);
		final Object message = in.readObject();
		return message;
	}
}
